package Fb::Objects;

use 5.006;
use strict;
use warnings;
use Data::Dumper;
use Fb::Out qw(out);
use Fb::Data qw(align human_readable $MAXINT64 timeStamp outputTable);

require Exporter;
our @ISA=qw(Exporter);
our @EXPORT_OK=qw(
	renderKeys
	reprSingleProperty
	propertyPath2Value
	propertyValueHr
	objectPropertiesHr
	postProcessValue
	getObjectTableLine
	renderObjects
	getObjects
	sortObjects
	objectsEqualOrUndef
);	

=item $text=propertyPath2Value($object,$path,%options)

Converts a given property path to a text. The method will do object lookups along the way. Objects
will be converted to 'short' reps. I<$text> is always a defined scalar.

Possible I<%options> are:

=over 4

=item B<scalar> Only handle scalar properties, abort with error on others. This is used for resolving sorting keys into values.

Example I<$path>s:

=over 4

=item B<group//managers:expand//id/name:align=5>

=item B<group//managers:separator=,:r=short:align=-3>

=back

=cut

sub propertyPath2Value {
	my ($obj,$path,%options)=@_;
	my @path=split /\/\//,$path;
	my $component=shift @path;
	my @options;
	if ( $component =~ /^([a-z0-9\/]+):(.*)$/ ) {
		$component=$1;
		@options=split /:/,$2;
	}
	my $value=$obj->get($component);
	my $fieldmeta=$obj->listKeys()->{$component};
	my $type=$fieldmeta->[2];

	my $undef='-'; # What to print for an undefined object
	if ( $options{scalar} ) {
		$undef=undef;
	}
	my $separator=' '; # What to separate multiple objects in a cell with
	my $process; # Do post processing
	my $repr='unique'; # Select repr value of an object
	my $align=0; # Align an object in a table cell
	my $expand=0; # Try to get multiple objects via $obj->expand() when encountering an object

	for my $option (@options) {
		if ( $option =~ /^(?:a|align)=(-?[0-9]+)$/ ) {
			if ( $options{scalar} ) {
				out('e',"Option 'a' is not allowed here.");
				return undef;
			}
			$align=$1;
		} elsif ( $option =~ /^(?:f|p|process|format)=(.*)$/ ) {
			$process=$1;
		} elsif ( $option =~ /^r=(\S+)$/ ) {
			if ( $options{scalar} ) {
				out('e',"Option 'r' is not allowed here.");
				return undef;
			}
			$repr=$1;
		} elsif ( $option =~ /^undef=(.*)$/ ) {
			$undef=$1;
		} elsif ( $option =~ /^separator=(.*)$/ ) {
			if ( $options{scalar} ) {
				out('e',"Option 'separator' is not allowed here.");
				return undef;
			}
			$separator=$1;
		} elsif ( $option eq 'expand' ) {
			if ( $options{scalar} ) {
				out('e',"Option 'expand' is not allowed here.");
				return undef;
			}
			$expand=1;
		} else {
			out('w',"Invalid property option '$option', ignored");
		}
	}
	if ( $options{scalar} && $type != 1 && $type != 4 ) {
		out('e',"Invalid type '$type' of field '$component' while looking for a scalar");
		return undef;
	}
	if ( $type == 1 ) {
		if ( defined($value) ) {
			postProcessValue(\$value,$process);
		} else {
			$value=$undef;
		}
	} elsif ( $type == 2 ) {
		$value=join($separator,grep { defined($_) } map { postProcessValue(\$_,$process);$_ } @{$value});
	} elsif ( $type == 3 ) {
		$value=join($separator,grep { defined($_) } map { postProcessValue(\$_,$process);$_ } sort keys %{$value});
	} elsif ( $type == 4 ) {
		my $expanded=[$value];
		if ( ! defined($value) ) {
			return align($undef,$align);
		}
		if ( $expand ) {
			$expanded=$value->expand();
		}
		$value=join($separator,map {
			if ( @path ) {
				# more '//'-components, scan deeper
				propertyPath2Value($_,join('//',@path));
			} else {
				$value->repr($repr);
			}
		} @{$expanded});
	}
	if ( $options{scalar} ) {
		return $value;
	}
	
	if ( $type == 5 || $type == 6 ) {
		if ( @{$value} == 0 ) {
			return align($undef,$align);
		}
		my $expanded=[];
		if ( $expand ) {
			for my $value_ (@{$value}) {
				my $expanded_=$value->expand();
				push(@{$expanded},@{$expanded_});
			}
		} else {
			$expanded=$value;
		}
		$value=join($separator,map {
			if (@path) {
				propertyPath2Value($_,join('//',@path));
			} else {
				$_->repr('short');
			}
		} @{$expanded});
	} elsif ( $type == 7 ) {
		if ( defined($value) ) {
			$value="(binary_data)";
		} else {
			$value=$undef;
		}
	}
	return align($value,$align);
}

sub getObjectTableLine {
	my ($obj,$format)=@_;
	my @paths=split /\|/,$format;
	return join('	',map { propertyPath2Value($obj,$_) } @paths);
}

sub objectPropertiesHr {
	my ($obj,%options)=@_;
	my $text='';
	my $keys=$obj->listKeys();
	my $unique=$obj->get('id/unique');
	for my $key (sort keys %{$keys}) {
		$text .= "$unique/$key " . propertyPath2Value($obj,$key) . "\n";
	}
	return $text;
}

sub reprSingleProperty {
	my ($obj,$key)=@_;
	my $type=$obj->listKeys()->{$key}->[2];
	if ( ! defined($type) ) {
		return '(invalid_field)';
	}
	my $value=$obj->get($key);
	return propertyValueHr($type,$value);
}


sub propertyValueHr {
	my ($type,$value,%options)=@_;
	my $objectformat='short';
	if ( $options{'format'} ) {
		$objectformat=$options{'format'};
	}
	if ( $type == 0 ) {
		if ( defined($value) ) {
			return $value;
		} else {
			return '?';
		}
	} elsif ( $type == 1 ) {
		if ( defined($value) ) {
			return $value;
		}
	} elsif ( $type == 2 ) {
		return join(' ',@{$value});
	} elsif ( $type == 3 ) {
		return join(' ',sort keys %{$value});
	} elsif ( $type == 4 ) {
		if ( $value ) {
			return $value->repr($objectformat);
		} else {
			return "-";
		}
	} elsif ( $type == 5 ) {
		my $sep=', ';
		if ( $objectformat eq 'unique' ) {
			$sep=' ';
		}
		return join($sep,map { $_->repr($objectformat) } @{$value});
	} elsif ( $type == 6 ) {
		my $sep=', ';
		if ( $objectformat eq 'unique' ) {
			$sep=' ';
		}
		return join($sep,map { $_->repr($objectformat) } sort { $a->get('id/unique') cmp $b->get('id/unique') } @{$value});
	} elsif ( $type == 7 ) {
		if ( defined($value) ) {
			return "(binary_data)";
		} else {
			return "-";
		}
	} else {
		return "?";
	}
}


=item \@sorted=sortObjects($objects,$path)

Sorts the list of objects in either I<@{$objects}> or in I<%{$objects}>
according to the object field I<$field>. The field's sorting policy is
derived from an arbitrary object. I<$field> being "B<->" will
make sortObjects() return the objects without sorting.

I<$field> is a regular object property and might be in the form of
 I<$actualfield>:I<option>{:I<option>}
Possible I<options> are:

=over 4

=item B<r> Reverse sort order

=back

=cut

sub sortObjects {
	my ($objects,$path)=@_;
	my @unsorted;
	if ( ref($objects) eq 'ARRAY' ) {
		@unsorted=@{$objects};
	} else {
		@unsorted=values %{$objects};
	}
	if ( $path eq '-' ) {
		return \@unsorted;
	}
	if ( ! @unsorted ) {
		return [];
	}
	
	my %sorthash;
	my $numeric=1;
	for my $obj (@unsorted) {
		my $sortvalue=propertyPath2Value($obj,$path,scalar => 1);
		$sorthash{$obj}=$sortvalue;
		if ( defined($sortvalue) && $sortvalue !~ /^[-0-9]+$/ ) {
			$numeric=0;
		}
	}
	my @sorted=sort {
		my $a_=$sorthash{$a};my $b_=$sorthash{$b};
		if ( ! defined($a_) and ! defined($b_) ) {
			0;
		} elsif ( defined($a_) and ! defined($b_) ) {
			1;
		} elsif ( ! defined($a_) and defined($b_) ) {
			-1;
		} elsif ( $numeric ) {
			($a_ <=> $b_);
		} else {
			($a_ cmp $b_);
		}
	} @unsorted;
	return \@sorted;
}

=item $result=renderObjects(\@objects,$outputformat,$sortorder,%options)

This function will print objects to STDOUT using the repr()-output mode
I<$outputformat> on all of them. They will be printed in order
according to objectproperty I<$sortorder>. Possible I<%options> are:

=over 4

=item B<tabfilter> Try to align tables in I<$outputformat> = B<table>I<...> .

I<$result> will return B<1>, if no error occured and B<0>, if there was a
problem (sorting or printing).

=item B<returntext> Return the text instead of printing it to STDOUT.

=back

I<$result> will contain:

=over 4

=item B<1>, if the objects were rendered successfully and I<$options{returntext}> is something false.

=item B<0>, if the objects were not rendered successfully and I<$options{returntext> is something false.

=item I<rendered objects> - the text returned by the object's I<repr()> method, if the rendering was successful and I<$options{returntext}> is something true.

=item I<undef>, if the objects were not rendered successfully and I<$options{returntext}> is something true.

=back

=cut

sub renderObjects {
	my ($objects,$outputformat,$sortorder,%options)=@_;
	my $sorted=sortObjects($objects,$sortorder) || return 0;
	my $success=1;
	my $text='';
	if ( $outputformat =~ /^table:(.*)/ ) {
		my @paths=split /\|/,$1;
		my @table;
		for my $obj (@{$sorted}) {
			push(@table,[map { propertyPath2Value($obj,$_) } @paths]);
		}
		if ( $options{tabfilter} ) {
			$text=outputTable(\@table);
		} else {
			for my $l (@table) {
				$text .= join("	",@{$l}) . "\n";
			}
		}
	} else {
		for my $obj (@{$sorted}) {
			my $s=$obj->repr($outputformat,%options);
			chomp($s);
			$text .= $s . "\n";
		}
	}
	if ( $options{returntext} ) {
		return $text;
	} else {
		print $text;
		return 1;
	}
}

sub renderKeys {
	my ($keys,%options)=@_;
	my $t='';
	my %resulttypes=(
		1 => "Scalar value",
		2 => "List of values",
		3 => "Set of values",
		4 => "Object",
		5 => "List of objects",
		6 => "Unordered list of objects",
		7 => "Binary data",
	);
	my $maxlength=0;
	for my $id (keys %{$keys}) {
		if ( length($id) > $maxlength ) {
			$maxlength=length($id);
		}
	}
	for my $id (sort keys %{$keys}) {
		my $r=$keys->{$id};
		my $writable='  ';
		if ( $r->[0] & 2 ) {
			substr($writable,0,1)='=';
		}
		if ( $r->[0] & 4 ) {
			substr($writable,1,1)='+';
		}
		my $space=' ' x ($maxlength-length($id));
		$t .= "$id$writable$space $r->[1]\n";
		if ( $options{mode} ) {
			my $result='?';my $comp='?';
			if ( defined($resulttypes{$r->[2]}) ) {
				$result=$resulttypes{$r->[2]};
			}
			if ( ref($r->[3]) eq 'ARRAY' ) {
				$comp="Objectfield '$r->[3]->[0]'";
			} elsif ( $r->[3] == 0 ) {
				$comp="(not_sortable)";
			} elsif ( $r->[3] == 1 ) {
				$comp="string_sort";
			} elsif ( $r->[3] == 2 ) {
				$comp="numeric_sort";
			}
			$t .= " " x $maxlength . "   result_type=$result, comparision_mode=$comp\n";
		}
	}
	return $t;
}



=item $success=postProcessValue(\$value,$mode)

Change a raw value into something different before outputting. This is
usually used to make a object property human readable. Possible I<$mode>s are:

=over 4

=item B<date>

Convert an epoch value into a date in style DD.MM.YYYY.

=item B<time>

Convert an epoch value into a timestamp in style YYYYMMDD-hhmmss.

=item B<tdiff>

Convert an epoch into an HR representation of the difference between the
value and "now".

=item B<memory>

Convert a value into a memory representation (1024 -> 1ki).

=item B<integer>

A unit prefixes (1000 -> 1000k).

=item B<text>

Convert a (potentially multiline) text into a single line text with quoted line terminators.

=item B<grep=>I<PCRE>

Filters the given value, returns the value unmodified, if it matches I<PCRE>.

=back

=cut

sub postProcessValue {
	my ($v,$rule)=@_;
	if ( ! defined($rule) ) {
		if ( ! defined($$v) ) {
			$$v='';
		}
		return 1;
	} elsif ( $rule =~ /^(?:epoch2date|date)(?:=(.*))?$/ ) {
		my $format=$1 || 'YYYYMMDD';
		if ( ! defined($$v) ) {
			$$v='?';
		} elsif ( $$v  == $MAXINT64 ) {
			$$v='-';
		} else {
			$$v=timeStamp($format,$$v);
		}
		return 1;
	} elsif ( $rule eq 'epoch2time' || $rule eq 'time' ) {
		if ( ! defined($$v) ) {
			$$v='?';
		} elsif ( $$v  == $MAXINT64 ) {
			$$v='-';
		} else {
			$$v=timeStamp('YYYYMMDD-hhmmss',$$v);
		}
		return 1;
	} elsif ( $rule eq 'timeinterval' ) {
		$$v=human_readable($$v,'timeinterval');
		return 1;
	} elsif ( $rule eq 'epoch2diff' || $rule eq 'tdiff' ) {
		if ( ! defined($$v) ) {
			$$v='?';
		} elsif ( $$v  == $MAXINT64 ) {
			$$v='-';
		} else {
			my $time=time();
			if ( $$v > time() ) {
				$$v='+' . human_readable($$v - $time,'timeinterval');
			} else {
				$$v=human_readable($time-$$v,'timeinterval');
			}
		}
		return 1;
	} elsif ( $rule eq 'memory' ) {
		if ( ! defined($$v) ) {
			$$v='?';
		}
		$$v=human_readable($$v,'memory');
		return 1;
	} elsif ( $rule eq 'integer' ) {
		if ( ! defined($$v) ) {
			$$v='?';
		}
		$$v=human_readable($$v,'integer');
	} elsif ( $rule eq 'text' ) {
		if ( ! defined($$v) ) {
			$$v='';
		} else {
			$$v =~ s/\\n/\n/gs;
			$$v =~ s/\\\\/\\/gs;
		}
		return 1;
	} elsif ( $rule =~ /^grep=(.*)$/ ) {
		if ( $$v !~ /^$1$/ ) {
			$$v=undef;
		}
	} else {
		out('e',"Unknown object field postprocess rule '$rule'");
		return 0;
	}
}

=item \%objectcollection=getObjects($class,\&getallhook,%options);


=cut

sub getObjects {
	my ($class,$getallhook,%options)=@_;
	my $objects=[];my $invalid=[];
	my $args=$options{args};
	my $success=1;
	if ( $options{all} ) {
		if ( $options{filter} ) {
			out('e',"--all and --filter cannot be used together");
			exit 3;
		}
		if ( @{$args} ) {
			out('e',"--all cannot be used when ids are given as parameters");
			exit 3;
		}
		$objects=&{$getallhook}(format => 'arrayref');
		if ( ref($objects) eq 'HASH' ) {
			# I want Python and cool per-class container-objects :-(
			$objects=[values %{$objects}];
		}
	} elsif ( $options{filter} ) {
		$objects=&{$getallhook}(format => 'arrayref',filter => $args);
		if ( ref($objects) eq 'HASH' ) {
			# I want Python and cool per-class container-objects :-(
			$objects=[values %{$objects}];
		}
	} else {
		my %createoptions;
		if ( $options{createoptions} ) {
			%createoptions=%{$options{createoptions}};
		}
		my $createkey='unique';
		if ( exists($options{createkey}) ) {
			$createkey=$options{createkey};
		}
		for my $arg (@{$args}) {
			my $obj;
			if ( defined($createkey) ) {
				eval "\$obj=${class}->new($createkey => \$arg,%createoptions);";
			} else {
				eval "\$obj=${class}->new(\$arg,%createoptions);";
			}
			if ( $@ ) {
				print STDERR $@;
			}
			if ( defined($obj) ) {
				push(@{$objects},$obj);
			} else {
				push(@{$invalid},$arg);
				$success=0;
			}
		}
	}
	if ( ! $success and ! $options{ignoreinvalid} ) {
		exit 1;
	}
	return {objects => $objects,invalid => $invalid,foundinvalid => int(! $success),onlyvalid => $success};
}


=item $result=objectsEqualOrUndef($object1,$object2)

Returns B<1>, if the given objects are identical (which means, their
'id/unique' property is the same) or if both objects are B<undef>.
Otherwise B<0> is returned.

=cut

sub objectsEqualOrUndef {
	my ($o1,$o2)=@_;
	if ( defined($o1) and defined($o2) ) {
		return int($o1->get('id/unique') eq $o2->get('id/unique'))
	} elsif ( ! defined($o1) and ! defined($o2) ) {
		return 1;
	} else {
		return 0;
	}
}

1;
