package Fb::Dialog;

use 5.006;
use strict;
use warnings;
use Carp;
use Data::Dumper;

my $have_lib_term_readpassword=1;
eval "use Term::ReadPassword;";
if ( $@ ) {
	$have_lib_term_readpassword=0;
}

use Fb::Out qw($VERBOSE isTty isX out);
use Fb::IO qw(iaexec $iaeexitcode);
use Fb::Data qw(check_passwd bash_quote $PASSWD_SPECIAL_CHARS);

require Exporter;

our @ISA=qw(Exporter);
our $VERSION='1.0';
our @EXPORT_OK=qw(
	message
	input
	inputPassword
	menu
	getLinecount
	$DIALOGCMD
	$DIALOG_MAXX
	$DIALOG_MAXY
);

our $DIALOGCMD="dialog --output-fd 3 3>&1 >&2";
our $DIALOG_MAXX=undef;
our $DIALOG_MAXY=undef;

=head1 NAME

AFS::Instant::Dialog - Stardard Dialogs for InstantAFS Skripts

=head1 SYNOPSIS

use AFS::Instant::Dialog;

my $password=input_password($title,$text);

=head1 REQUIRES

dialog (the shell tool)

=head1 DESCRIPTION

Simple perl bindings for showing interactive dialogs in perl scripts.

=head1 FUNCTIONS

=over 4

=item out_dialog($type,$text{,$text2})

Outputs a message to the user (using ncurses). It is to be used
to re-route all output of a running script to ncurses dialogs.

use Fb::Out qw(setOutChannel);
use Fb::Dialog qw(out_dialog);
setOutChannel(\&out_dialog);
out('i',"This is a message);

=cut

sub out_dialog {
	my ($type,$text)=@_;
	if ( $type eq 'i' ) {
		message("Information",$text);
	} elsif ( $type eq 'v' ) {
		if ( $Fb::Out::VERBOSE ) {
			message("Information",$text);
		}
	} elsif ( $type eq 'd' ) {
		if ( $Fb::Out::VERBOSE > 1 ) {
			message("Debug message",$text);
		}
	} elsif ( $type eq 'dd' ) {
		if ( $Fb::Out::VERBOSE > 2 ) {
			message("Debug message (extended)",$text);
		}
	} elsif ( $type eq 'e' ) {
		message("Error",$text);
	} elsif ( $type eq 'w' ) {
		message("Warning",$text);
	}
}




=item $password=inputPassword($title,$text,%options)

Asks the user to enter a password. Possible I<%options> are:

=over 4

=item B<check>

Test the password for strength. Only return a password string
if it's strong enough.

=item B<confirm>>

Ask for password twice. Only return a password, if the two match.

=back

=cut

sub inputPassword {
	my ($title,$text,%options)=@_;
	my $titleq=quotemeta($title);
	my $textq=quotemeta($text);
	my $cmd='';
	my $password;
	while ( 1 ) {
		$password=inputPasswordSimple($title,$text,%options);
		if ( ! defined($password) ) {
			return undef;
		}
		if ( ! $options{check} || check_passwd($password) ) {
			last;
		} else {
			message('Error',"The given password is not strong enough or you used invalid characters. You may use Alphanumeric characters and those: $PASSWD_SPECIAL_CHARS.");
		}
	}
	if ( ! $options{confirm} ) {
		return $password;
	} else {
		my $temp=inputPasswordSimple($title,$text,%options);
		if ( $temp eq $password ) {
			return $password;
		}
		message('Error',"Passwords didn't match.");
		return undef;
	}
}

sub inputPasswordSimple {
	my ($title,$text,%options)=@_;
	my $password;
	if ( Fb::Out::isTty() && ! ( $options{usex} && Fb::Out::isX() ) ) {
		# This is a terminal and no X-usage is forced (and possible)
		if ( $options{usencurses} ) {
			my $titleq=quotemeta($title);
			my $textq=quotemeta($text); 
			my $cmd="$DIALOGCMD --title $titleq --passwordbox $textq 15 70";
			$password=`$cmd`;chomp($password);
			if ( ( $? >> 8 ) != 0 ) {
				return undef;
			}
		} elsif ( $have_lib_term_readpassword ) {
			$Term::ReadPassword::USE_STARS=1;
			$Term::ReadPassword::USE_STARS=1;
			$password=read_password($title,0,1);
		} else {
			### There's most likely a better way to turn off echo...
			system("stty -echo");
			$|=1;
			print "$title:";
			$password=<STDIN>;chomp($password);
			system("stty echo");
			if ( $password eq '' ) {
				return undef;
			}
		}
	} elsif ( Fb::Out::isX() ) {
		# Found X-Display. Use graphical password prompt
		if ( -x "/usr/bin/ssh-askpass" ) {
			my $titleq=quotemeta($title);
			$password=iaexec(
				['/usr/bin/ssh-askpass',$title],
				'readpipe',
				description => "Asking user for password via X",
				format => 'string',
			);
			if ( ! defined($password) ) {
				return undef;
			}
			chomp($password);
		}
	} else {
		# No terminal - read password from stdin
		print $title . "\n";
		my $password=<STDIN>;chomp($password);
	}
	return $password;
}

=item message($title,$text)

Shows the message I<$text> using 'dialog'. The user must hit 'return'
to continue.

=cut

sub message {
	my ($title,$text)=@_;
	my $titleq=quotemeta($title);
	my $textq=quotemeta($text);
	iaexec("$DIALOGCMD --title $titleq --msgbox $textq 20 70",'exec');
	return 1;
}




=item $result=input($title,$text,$default)

This function shows an input-box (title is I<$title>, text shown is I<$text>).
The string contained defaults to I<$default>.
I<$result> contains the entered (or modified) content of the input-box.
It's not defined if the user aborted.

=cut

sub input {
	my ($title,$text,$default)=@_;
	getTermSize();
	my $titleq=bash_quote($title);
	my $textq=bash_quote($text);
	if ( ! defined($default) ) {
		$default='';
	}
	my $defaultq=bash_quote($default);
	my $cmd="$DIALOGCMD --title $titleq --inputbox $textq $DIALOG_MAXY $DIALOG_MAXX $defaultq";
	my $input=iaexec($cmd,'readpipe',stderr => 'show',format => 'string');
	if ( ! defined($input) ) {
		return undef;
	}
	chomp($input);
	return $input;
}


=item ($result,$item)=menu($title,$text,\@entries,\@buttons,%options)

Shows a ncurses-menu with a given I<$title>, a given I<$text> , given menu
I<@entries> and one or more I<@buttons>.
Examples for I<@entries> and I<@buttons>:

\@entries=[
 ['item1', 'This is item 1'],
 [$item2,$text_to_be_shown]
];

\@entries=[
 ['check1', 'This is an unchecked checkbox',0],
 ['check2', 'This is a checked one',1]
];

\@buttons=[
 ' Select ',
 ' Abort '
 ' New ',
];

I<$result> : The first Button is always the selection button - returning B<1> when pressed.
 The second one is the exit-button, whose return value is B<0>.
 Pressing Esc also leads to I<$result> == B<0>.
 Other buttons (there can be 0, 1 or 2 of them) will return B<2> or B<3> when pushed.

I<$item> : the key of the selected menu item. I<$item> is B<undef> whenever the last button
 was pushed

These I<%options> can be given:

=over 4

=item B<size_x>

X-Size of the dialog window (Default: maximum)

=item B<size_y>

Y-Size of the dialog window (Default: maximum)

=item B<menu_y> 

Height of the menubox inside the dialog. Default: As much
as possible while displaying the whole given text - but not
less than 5 (3 entries shown at once)).

=item B<mode>

The kind of dialog to be used - currently either 'B<menu>' for a simple
list menu or 'B<checklist>' for a checkbox list dialog. Default: B<menu>.

=item B<default> - Key of the default item to be pre-selected.

=back

=cut

sub menu {
	my ($title,$text,$entries,$buttons,%conf)=@_;
	getTermSize();
	my $mode='menu';
	if ( defined($conf{'mode'}) ) {
		$mode=$conf{'mode'};
	}
	my $size_x=$DIALOG_MAXX;
	if ( defined($conf{'size_x'}) ) {
		$size_x=$conf{'size_x'};
	}
	my $size_y=$DIALOG_MAXY;
	if ( defined($conf{'size_y'}) ) {
		$size_y=$conf{'size_y'};
	}
	my $text_y=getLinecount($text,$size_x-4);
	my $menu_y=$size_y-6-$text_y;
	if ( defined($conf{'menu_y'}) ) {
		$menu_y=$conf{'menu_y'};
		if ( ( $text_y + $menu_y + 7 ) <= $DIALOG_MAXY ) {
			$size_y=$text_y + $menu_y + 7;
		}
	} else {
		if ( ! defined($conf{'size_y'}) ) {
			if ( $menu_y > @{$entries} ) {
				$size_y -= ( $menu_y - @{$entries} );
				$menu_y=@{$entries};
			}
		} else {
			if ( $menu_y < 3 ) {
				$menu_y=3;
			}
		}
	}
	my $cmd="$DIALOGCMD --title " . bash_quote($title) . " ";
	if ( @{$buttons} == 0 ) {
		message('Error',"menu() called but no button given, internal error.");
		return (0,undef);
	} elsif ( @{$buttons} > 4 ) {
		message('Error',"menu() called but more than 4 buttons given, internal error.");
		return (0,undef);	
	} elsif ( @{$buttons} == 1 ) {
		$cmd .= "--no-cancel ";
	} else {
		$cmd .= "--cancel-label " . bash_quote($buttons->[1]) . " ";
		if ( @{$buttons} >= 3 ) {
			$cmd .= "--help-button --help-label " . bash_quote($buttons->[2]) . " ";
			if ( @{$buttons} == 4 ) {
				$cmd .= "--extra-button --extra-label " . bash_quote($buttons->[3]) . " ";
			}
		}
	}
	$cmd .= "--ok-label " . bash_quote($buttons->[0]) . " ";
	if ( defined($conf{default}) ) {
		$cmd .= "--default-item " . bash_quote($conf{default}) . " ";
	}
	if ( $mode eq 'menu' ) {
		$cmd .= '--menu ';
	} elsif ( $mode eq 'checklist' ) {
		$cmd .= '--checklist ';
	}
	$cmd .= bash_quote($text) . " $size_y $size_x $menu_y ";
	if ( ! @{$entries} ) {
		out('e',"menu() called without menu entries, internal error.");
		exit 1;
	}
	foreach my $entry ( @{$entries} ) {
		$cmd .= bash_quote($entry->[0]) . " " . bash_quote($entry->[1]) . " ";
		if ( $mode eq 'checklist' ) {
			if ( $entry->[2] ) {
				$cmd .= 'on ';
			} else {
				$cmd .= 'off ';
			}
		}
	}
	my $item;
	my $result=iaexec($cmd,'readpipe',stderr => 'show',format => \$item);chomp($item);
	my $exitcode=$iaeexitcode >> 8;
	if ( ( $exitcode == 0 ) or ( $exitcode == 3 ) ) {
		if ( $exitcode == 0 ) { $exitcode=1; }
		if ( $mode eq 'checklist' ) {
			foreach my $entry (@{$entries}) {
				if ( $item =~ /\"?$entry->[0]\"?/ ) {
					$entry->[2]=1;
				} else {
					$entry->[2]=0;
				}
			}
			return ($exitcode,undef);
		} else {
			return ($exitcode,$item);
		}
	} elsif ( ( $exitcode == 1 ) or ( $exitcode == 255 ) ) {
		return (0,undef)
	} elsif ( $exitcode == 2 ) {
		if ( $item =~ /^HELP (\S+)$/ ) {
			return (2,$1);
		} else {
			return (2,0);
		}
	} else {
		out('e',"Unexpected return code received from dialog: '$exitcode'");
		out('i',"command was: '$cmd'");
		return ((undef),(undef));
	}
}




=back

=head2 Helper functions

=over 4

=item $ok=getTermSize()

Reads maximum possible dialog-screen-dimensions by calling
'dialog --print-maxsize'

=cut

sub getTermSize {
	if ( defined($DIALOG_MAXX) && defined($DIALOG_MAXX) ) {
		return 1;
	}
	if ( open(X,"$DIALOGCMD --print-maxsize|") ) {
		my $l=<X>;
		chomp($l);
		if ( $l =~ /^MaxSize: ([0-9]+), ([0-9]+)$/ ) {
			$DIALOG_MAXX=$2;
			$DIALOG_MAXY=$1;
		}
		close(X);
	}
	if ( defined($DIALOG_MAXX) ) {
		return 1;
	} else {
		out('e',"Unable to get terminal dimensions from dialog.");
		exit 1;
	}
}


=item $lines=getLinecount($text,$width)

Estimates the number of lines I<$text> would use when outputtet
in a 'dialog'-window of width I<$width>.

=cut

sub getLinecount {
	my ($text,$width)=@_;
	$width -= 7;
	my $space=0;
	my $hardwrap=1;
	my $lines=0;
	my $line=0;
	for(my $n=0;$n<length($text);$n++) {
		my $c=substr($text,$n,1);
		if ( $c eq "\n" ) {
			$line=0;$space=0;$lines++;$hardwrap=1;
		} elsif ( $c =~ /^\s$/ ) {
			if ( $hardwrap ) {
				if ( $line < $width ) {
					$space=1;$line++;
				} else {
					$lines++;$line=0;$hardwrap=0;$space=0;
				}
			} else {
				if ( $space < 2 ) {
					if ( $line != 0 ) {
						if ( $line < $width ) {
							$space++;
							$line++;
						} else {
							$lines++;$line=0;$space=0;$hardwrap=0;
						}
					}
				} else {
					if ( $line == $width ) {
						$lines++;$line=0;$space=0;$hardwrap=0;
					}
				}
			}
		} else {
			$space=0;
			if ( $line < $width ) {
				$line++;
				if ( $lines == 0 ) {
					$hardwrap=0;
				}
			} else {
				$hardwrap=0;$lines++;$line=0;
			}
		}
	}
	$lines++;
	return $lines;
}


=back

=head1 AUTHOR

Frank Burkhardt <burk@cbs.mpg.de>

=cut

1;
__END__
