package Fb::Out;

use 5.006;
use strict;
use warnings;
use Data::Dumper;

use Fb::Data qw($outmethod wrapLine);
use Getopt::Long;

require Exporter;
our @ISA=qw(Exporter);
our @EXPORT=qw(out);
our @EXPORT_OK=qw(
	out
	$outchannelname
	setOutChannel
	out_stderr
	out_stdout
	out_syslog
	out_xless
	$VERBOSE
	initOut
	getCommandBlock
	makeOptionsHash
	renderHelp
	isTty
	isX
);	

our $VERBOSE=0;
our $outchannelname; # To be imported by scripts
our $outchannel;
my $_outchannelset=0;

# Use libunix-syslog-perl if available
our $libsyslogperl=0;
eval "use Unix::Syslog;";
if ( $@ eq '' ) {   
	$libsyslogperl=1;
}


=item out($type,$text)

This method is the core of this module. It will show a given message I<$text> to
the user. These are the I<$type>s accepted:

=over 4

=item B<i>

Information - This message is always shown. It is used for non-critical information
and for explaination of preceeding errors or warnungs. It is also used for
what-if messages caused by dryrun operation.

=item B<w>

Warning. This message informs about non-critical (which means the program is able to handle
them without abort) problems.

=item B<e>

Error. This is a critical error message which will cause the running program to be interrupted.

=item B<v>

Verbose information. This message is only shown if I<$VERBOSE> is B<1> or higher.

=item B<d>

Debug message. This is only shown if I<$VERBOSE> is B<2> or higher (B<3> in case of B<dd> and so on).

=over 4

out() itself is just a frontend calling the currently active out-plugin.

=cut

sub out {
	return &{$outchannel}(@_);
}


=item $text=_out_make_lines($type,$text_raw)

This method prepares the I<$text> to be shown for simply text-oriented out-plugins.

=cut

sub _out_make_lines {
	my ($type,$text)=@_;
	$text =~ s/\x0d//g;
	my @text=split /\n/,$text;
	my $prefix;
	my $text2='';
	if ( $type eq 'i' ) {
		$prefix='I: ';
	} elsif ( $type eq 'v' ) {
		if ( ! $VERBOSE ) {
			return '';
		}
		$prefix="I: ";
	} elsif ( $type eq 'w' ) {
		$prefix="W: ";
	} elsif ( $type eq 'e' ) {
		$prefix="E: ";
	} elsif ( $type =~ /^d+$/ ) {
		if ( $VERBOSE <= length($type) ) {
			return '';
		}
		$prefix="D: ";
	} else {
		$text2="E: Unknown output message type '$type', internal error. Using '?'.\n";
		$prefix="?: ";
	}
	foreach my $line (@text) {
		$text2 .= "$prefix$line\n";
	}
	return $text2;
}


sub out_stderr {
	my ($type,$text_)=@_;
	my $text=_out_make_lines($type,$text_);
	print STDERR $text;
	return 1;
}

sub out_stdout {
	my ($type,$text_)=@_;
	my $text=_out_make_lines($type,$text_);
	print STDOUT $text;
	return 1;
}

my $syslog_open=0;
sub out_syslog {
	my ($type,$text)=@_;
	if ( $syslog_open == 0 ) {
		my $progname=$0;
		$progname =~ s/^(.*)\/([^\/]+)$/$2/;
		Unix::Syslog::openlog("$progname\[$$\]",0,24); # LOG_DAEMON
		$syslog_open=$$;
	} elsif ( $syslog_open != $$ ) {
		my $progname=$0;
		$progname =~ s/^(.*)\/([^\/]+)$/$2/;
		Unix::Syslog::closelog();
		Unix::Syslog::openlog("$progname\[$$\]",0,24); # LOG_DAEMON
		$syslog_open=$$;
	}
	$text =~ s/\x0d//g;
	my @text=split /\n/,$text;
	if ( $type eq 'i' ) {
		foreach my $line (@text) {
			Unix::Syslog::syslog(6,"%s",$line); # LOG_INFO
		}
	} elsif ( $type eq 'v' ) {
		if ( $VERBOSE ) {
			foreach my $line (@text) {
				Unix::Syslog::syslog(6,"%s",$line); # LOG_INFO
			}
		} else {
			return 0;
		}
	} elsif ( $type =~ /^d+$/ ) {
		if ( $VERBOSE > length($type) ) {
			foreach my $line (@text) {
				Unix::Syslog::syslog(7,"%s",$line); # LOG_DEBUG
			}
		} else {
			return 0;
		}
	} elsif ( $type eq 'w' ) {
		foreach my $line (@text) {
			Unix::Syslog::syslog(4,"%s",$line);
		}
	} elsif ( $type eq 'e' ) {
		foreach my $line (@text) {
			Unix::Syslog::syslog(3,"%s",$line);
		}
	} else {
		Unix::Syslog::syslog(3,"(unknown message): %s",$text);
	}
}

my $xless_pipe;
my $xless_pid;
sub out_x {
	my ($type,$text_)=@_;
	my $text=_out_make_lines($type,$text_);
	if ( ! $text ) {
		return 1;
	}
	if ( ! $xless_pid ) {
		my $readhandle;
		pipe($readhandle,$xless_pipe);
		$xless_pipe->autoflush(1);
		
		$xless_pid=fork();
		if ( ! $xless_pid ) {
			close($xless_pipe);
			open(STDIN,">&",$readhandle);
			open(STDOUT,"> /dev/null");
			open(STDERR,"> /dev/null");
			exec("/usr/bin/xless","-f","-title","Output of '$0'");
			exit 1;
		} else {
			close($readhandle);
		}
	}
	print $xless_pipe $text;
	return 1;
}

my $logfile_open;
my $logfile_name;
sub out_file {
	my ($type,$text_)=@_;
	my $text=_out_make_lines($type,$text_);
	if ( ! $text ) {
		return 1;
	}
	if ( ! $logfile_open ) {
		if ( ! open($logfile_open,">> $logfile_name") ) {
			out_stderr('e',"Unable to open logfile '$logfile_name'");
			return 0;
		}
	}
	print $logfile_open $text;
	return 1;
}

my $_outputchannel;

=item $result=initOut(%options)

This method's main purpos is to parse commandline parameters and to initialize the
info/warning/error reporting subsystem.

=cut

sub initOut {
	my %options=@_;
	my %getoptoptions=(
		'bundling' => 1,
		'no_getopt_compat' => 1,
	);
	if ( $options{getoptoptions} ) {
		for my $option (keys %{$options{getoptoptions}}) {
			if ( $options{getoptoptions}->{$option} ) {
				$getoptoptions{$option}=1;
			} else {
				delete $getoptoptions{$option};
			}
		}
	}
	Getopt::Long::Configure(sort keys %getoptoptions);
	my %cmdoptions=();
	if ( $options{cmdoptions} ) {
		%cmdoptions=%{$options{cmdoptions}};
	} elsif ( $options{blocks} ) {
		%cmdoptions=%{makeOptionsHash(blocks => $options{blocks})};
	}
	$cmdoptions{'verbose|v+'}=\$VERBOSE;
	$cmdoptions{'output-media|Y=s'}=\$_outputchannel,
	my $result=GetOptions(
		%cmdoptions,
	);
	if ( ! $result ) {
		out('e',"Invalid command line parameters given. You need --help .");
		return 0;
	}

	# In 'safemode',do no accept certain output channels via commandline
	if ( $options{safemode} && defined($_outputchannel) && $_outputchannel !~ /^(console|stdout)/ ) {
		$_outputchannel=undef;
	}
	# Render help automatically (if requested)
	if ( defined($options{help}) ) {
		if ( ${$options{help}} ) {
			exit ! renderHelp(%options);
		}
	}

	if ( defined($_outputchannel) ) {
		$result &= setOutChannel($_outputchannel);
	} elsif ( ! $_outchannelset ) {
		# Provide a default channel, if non was already set explicitely
		if ( ! isTty() && isX() ) {
			setOutChannel('x');
		} else {
			setOutChannel('console');
		}
	}
	# By default, signal-toggling of the debug level is enabled.
	if ( ! defined($options{togglesignal}) || $options{togglesignal} ) {
		$SIG{USR1}=\&debugToggleSignal;
	}
	return $result;
}


=item setOutChannel(\&out_procedure)

Defines a new output method to be used when I<out()> is called.
This could either be a method ref or an ID. Valid IDs are:
B<console>, B<stdout> and B<syslog>.

=cut

sub setOutChannel {
	my $new=shift;
	$_outchannelset=1;
	if ( $logfile_open ) {
		close($logfile_open);
		$logfile_name=undef;
	} elsif ( $xless_pid ) {
		sleep 5;
		kill(9,$xless_pid);
		$xless_pid=0;
	}
	if ( ref($new) eq 'CODE' ) {
		$outchannel=$new;
	} elsif ( ref($new) eq '' ) {
		if ( $new eq 'stderr' || $new eq 'console' ) {
			$outchannel=\&out_stderr;
			$outchannelname='stderr';
		} elsif ( $new eq 'stdout' ) {
			$outchannel=\&out_stdout;
			$outchannelname='stdout';
		} elsif ( $new eq 'x' ) {
			$outchannel=\&out_x;
			$outchannelname='x';
		} elsif ( $new eq 'syslog' ) {
			if ( $libsyslogperl ) {
				$outchannel=\&out_syslog;
				$outchannelname='syslog';
			} else {
				out('e',"setOutChannel(): Syslog unavalable");
				return 0;
			}
		} elsif ( $new =~ /^file:(.+)$/ ) {
			$outchannelname=$new;
			$logfile_name=$1;
			$outchannel=\&out_file;
		} else {
			print STDERR "E: setOutChannel(): Invalid channel ID\n";
			return 0;
		}
	} else {
		print "E: setOutChannel(): Invalid parameter format\n";
		return 0;
	}
	return 1;
}

=item \%block=getCommandBlock()

Returns a command block to be included to be included into a initOut() call.
This block contains option controlling L<Fb::Out>'s behaviour.

I<%block> is a command hash to be included into L<initOut()>'s B<blocks>
arrayref. It's guaranteed to contain something useful - no error testing
is necessary.

Possible I<%options> are:

=over 4

=item B<ftest> => B<1>

Act as if xless exists.

=back

=cut

sub getCommandBlock {
	my %options=@_;
	my @addchannels=();
	if ( $libsyslogperl ) {
		push(@addchannels,"syslog - send output to syslog");
	}
	if ( ( -x "/usr/bin/xless" && $ENV{'DISPLAY'} ) || $options{'ftest'} ) {
		push(@addchannels,"x - show output in a separate X-window");
	}
	my %block=(
		title => "FB::Out output options",
		parameters => [
			{name => 'verbose|v+',text => ["Be verbose. Use twice or more to increase verbosity"],var => \$VERBOSE},
			{name => 'output-media|Y=s',arg => 'M',text => [
				"Use output media M for error/debug/information output. Possible media are:",[
					'stderr - sent output to stderr',
					"stdout - send output to stderr",
					@addchannels,
					"file:FILE - send output to logfile FILE",
				],
			],var => \$_outputchannel},
		]
	);
	return \%block;
}

sub renderHelp {
	my (%options)=@_;
	my $width=80;
	if ( $options{width} ) {
		$width=$options{width};
	}
	my $optwidth=20;
	if ( $options{optwidth} ) {
		$optwidth=$options{optwidth};
	}
	my $commandline=$options{commandline};
	my $author=$options{author};
	my $digest=$options{digest};
	my $blocks=$options{blocks};
	if ( ! $commandline or ! $author or ! $digest or ! $blocks ) {
		Fb::Out::out('e',"renderHelp(): A required field is missing. Unable to display help.");
		if ( $options{test} ) {
			return undef;
		} else {
			return 0;
		}
	}
	my $text="Usage: $commandline\n\n";
	if ( defined($digest) ) {
		$text .= Fb::Data::wrapLine($digest,$width) . "\n";
	}
	my $bcount=0;
	for my $block (@{$blocks}) {
		$bcount++;
		if ( ! defined($block->{title}) ) {
			out('e',"renderHelp(): Parameter block #'$bcount' doesn't have a title");
			return 0;
		}
		$text .= "$block->{title}:\n\n";
		my $count=0;
		for my $param (@{$block->{parameters}}) {
			$count++;
			if ( ! $param->{name} ) {
				if ( scalar(keys %{$param}) != 0 ) {
					out('e',"renderHelp(): Parameter #'$count' in block '$block->{title}' doesn't have a name but is not empty");
					return 0;
				}
				$text .= "\n";
				next;
			}
			if ( $param->{name} =~ /^(\S+?)(\|\S+?)?(=[is]|\+)?$/ ) {
				if ( ! defined($param->{var}) ) {
					out('e',"renderHelp(): Missing variable pointer for option '$param->{name}'");
				}
				my ($long,$sec,$type)=($1,$2,$3);
				my $desc='     ';
				my @sec;
				if ( defined($sec) ) {
					# Make sure to have a one-char option at the first position
					@sec=sort { length($a) <=> length($b) } split /\|/,substr($sec,1,length($sec)-1);
					if ( length($sec[0]) == 1 ) {
						$desc=" -$sec[0], ";
						shift @sec;
					}
				}
				$desc .= "--$long";
				if ( defined($type) ) {
					if ( $type =~ /^=([is])/ ) {
						if ( defined($param->{arg}) ) {
							$desc .= "=$param->{arg}";
						} else {
							$desc .= "=$1";
						}
					}
				}
				for my $s (@sec) {
					if ( length($s) == 1 ) {
						$desc .= ", -$s";
					} else {
						$desc .= ", --$s";
					}
				}
				if ( length($desc) <= $optwidth ) {
					$text .= $desc . (" " x ($optwidth-length($desc))) . " ";
				} else {
					$text .= $desc . "\n" . (" " x ($optwidth+1));
				}
				my $prefixtype='abfprefix';
				if ( ! defined($param->{text}) ) {
					out('e',"renderHelp(): Text is missing for option '$param->{name}'");
					return 0;
				}
				for my $t (@{$param->{text}}) {
					if ( ref($t) eq '' ) {
						$text .= Fb::Data::wrapLine($t,$width,$prefixtype => (" " x ($optwidth+1)));
					} elsif ( ref($t) eq 'ARRAY' ) {
						for my $i (@{$t}) {
							$text .= Fb::Data::wrapLine($i,$width,abfprefix => (" " x ($optwidth+2)),fprefix => (" " x ($optwidth+1)));
						}
					} elsif ( ref($t) eq 'HASH' ) {
						my @lines;
						for my $key (sort keys %{$t}) {
							push(@lines,"$key - $t->{$key}");
						}
						for my $i (@lines) {
							$text .= Fb::Data::wrapLine($i,$width,abfprefix => (" " x ($optwidth+2)),fprefix => (" " x ($optwidth+1)));
						}
					}
					$prefixtype='prefix';
				}
			} else {
				out('e',"renderHelp(): Invalid param description '$param->{name}'");
				return 0;
			}
		}
		$text .= "\n";
	}
	if ( defined($options{examples}) ) {
		$text .= "Examples:\n\n";
		my $count=0;
		for my $e (@{$options{examples}}) {
			$count++;
			if ( ! defined($e->{command}) || ! defined($e->{text}) ) {
				out('e',"renderHelp(): Error in example description #'$count'");
				if ( $options{test} ) {
					return undef;
				} else {
					return 0;
				}
			}
			$text .= Fb::Data::wrapLine($e->{command},$width,prefix => ' ') . Fb::Data::wrapLine($e->{text},$width,prefix => '  ') . "\n";
		}
	}
	if ( $options{test} ) {
		return $text;
	} else {
		print $text;
		return 1;
	}
}

sub makeOptionsHash {
	my (%options)=@_;
	my %hash;
	for my $block (@{$options{blocks}}) {
		for my $param (@{$block->{parameters}}) {
			if ( $param->{name} ) {
				$hash{$param->{name}}=$param->{var};
			}
		}
	}
	return \%hash;
}


sub isTty {
	if ( -t STDIN ) {
		return 1;
	} else {
		return 0;
	}
}

sub isX {
	if ( defined($ENV{DISPLAY}) ) {
		return 1;
	} else {
		return 0;
	}
}


sub END {
	# Close open X-logoutput-window after a waiting time of 5 seconds
	if ( $xless_pid ) {
		out('i',"Process terminated, logwindow will close in 5 seconds");
		sleep 5;
		kill(9,$xless_pid);
	}
}

sub debugToggleSignal {
	if ( $VERBOSE >= 3 ) {
		$VERBOSE=0;
	} else {
		$VERBOSE++;
	}
	out('i',"Received SIGUSR1, set verbosity to '$VERBOSE'");
}

$outchannel=\&out_stderr;
$Fb::Data::outmethod=\&out;
1;
